﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections;

namespace Chapter06
{
    class Program
    {
        static void Main(string[] args)
        {
            // Call a sample method here.
        }

        static void Cast()
        {
            var list = new ArrayList();
            list.Add("Bart");
            list.Add("John");
            list.Add("Lisa");
            string bartIsBack = (string)list[0]; // Okay
            Console.WriteLine(bartIsBack);

            object o = 5;
            string s = (string)o; // Fail
            Console.WriteLine(s);
        }

        static void Is()
        {
            object o = "Bart";
            bool isString = o is string; // true
            bool isInt = o is int; // false

            Console.WriteLine(isString);
            Console.WriteLine(isInt);
        }

        static bool IsFresh_Is(Food item)
        {
            if (item is Fruit)
            {
                return IsFreshFruit((Fruit)item);
            }
            else if (item is Vegetable)
            {
                return IsFreshVegetable((Vegetable)item);
            }
            else
            {
                throw new InvalidOperationException("Unknown food type.");
            }
        }

        static void As()
        {
            object o = "Bart";
            string s = o as string;
            if (s != null)
            {
                Console.WriteLine(s);
            }
        }

        static bool IsFresh_As(Food item)
        {
            Fruit f;
            Vegetable v;
            if ((f = item as Fruit) != null)
            {
                return IsFreshFruit(f);
            }
            else if ((v = item as Vegetable) != null)
            {
                return IsFreshVegetable(v);
            }
            else
            {
                throw new InvalidOperationException("Unknown food type.");
            }
        }

        static bool IsFreshFruit(Fruit fruit)
        {
            throw new NotImplementedException();
        }

        static bool IsFreshVegetable(Vegetable vegetable)
        {
            throw new NotImplementedException();
        }

        static void Reflection()
        {
            DateTime now = DateTime.Now;
            Type t = now.GetType();
            foreach (var prop in t.GetProperties())
            {
                Console.WriteLine(prop.Name);
            }
        }

        static void Default()
        {
            var sample = new Sample<int>(42);
            sample.Print();

            sample.Reset();
            sample.Print();

            int zero = default(int);
            string nothing = default(string);
            DateTime longAgo = default(DateTime); // 01/01/0001 12:00:00 AM
            Console.WriteLine(zero);
            Console.WriteLine(nothing);
            Console.WriteLine(longAgo);
        }

        static void New()
        {
            var oneHourAndAHalf = new TimeSpan(1, 30, 0);
            Console.WriteLine(oneHourAndAHalf);

            var numbers = new List<int>();
            Console.WriteLine(numbers.Count);

            var tenStars = new string('*', 10);
            Console.WriteLine(tenStars);
        }

        static void ObjectInitializers()
        {
            var me = new Person("Bart") { Age = 26 };
            Console.WriteLine(me.Name + " is " + me.Age);
        }

        static void CollectionInitializers()
        {
            var primes = new List<int> { 2, 3, 5, 7 };
            foreach (var prime in primes)
                Console.WriteLine(prime);

            var ages = new Dictionary<string, int> { { "Bart", 26 }, { "John", 62 } };
            foreach (var name in ages.Keys)
                Console.WriteLine(name + " is " + ages[name]);
        }

        static void MethodInvocation()
        {
            Console.WriteLine("Hello");

            var person = new Person("Bart");
            string s = person.ToString();
            Console.WriteLine(s);

            var calc = new Calculator();
            int three = calc.Add(1, 2);
            Console.WriteLine(three);
        }

        static void DelegateInvocation()
        {
            var calc = new Calculator();
            Console.WriteLine(calc.Ext(new BinOp(Hypotenuse), 3, 4));
        }

        static int Hypotenuse(int a, int b)
        {
            return (int)Math.Sqrt(a * a + b * b);
        }

        static void AnonymousMethods()
        {
            var calc = new Calculator();
            Console.WriteLine(calc.Ext(
                delegate (int a, int b) { return (int)Math.Sqrt(a * a + b * b); },
                3, 4));
        }

        static void LambdaExpressions()
        {
            var calc = new Calculator();
            Console.WriteLine(calc.Ext((a, b) => (int)Math.Sqrt(a * a + b * b), 3, 4));
        }

        static void ElementAccess()
        {
            // Arrays
            {
                int[] primes = new int[] { 2, 3, 5, 7 };
                int five = primes[2];
                Console.WriteLine(five);
            }

            // Indexers
            {
                var primes = new List<int> { 2, 3, 5, 7 };
                // Index-based access just as with arrays. Lists are flexible arrays after all.
                int five = primes[2];
                Console.WriteLine(five);

                var ages = new Dictionary<string, int> { { "Bart", 26 }, { "John", 62 } };
                // Indexers can also use a different type.
                // Dictionary<K,T> uses an index of type K, for the dictionary’s key type.
                int bartAge = ages["Bart"];
                Console.WriteLine("Bart is " + bartAge);

                // We can add or replace key/value pairs in the dictionary as follows.
                // This uses a setter method underneath.
                ages["Lisa"] = 8;
                Console.WriteLine("Lisa is " + ages["Lisa"]);
            }
        }
    }

    class Food
    {
    }

    class Fruit : Food
    {
    }

    class Apple : Fruit
    {
    }

    class Kiwi : Fruit
    {
    }

    class Vegetable : Food
    {
    }

    class Cookie : Food
    {
    }

    class Sample<T>
    {
        private T _data;

        public Sample(T data)
        {
            _data = data;
        }
        
        public void Reset()
        {
            _data = default(T);
        }

        public void Print()
        {
            Console.WriteLine(_data);
        }
    }

    class Person
    {
        public Person(string name)
        {
            Name = name;
        }

        public string Name { get; set; }
        public int Age { get; set; }
    }

    class Calculator
    {
        public int Add(int a, int b) { return a + b; }
        public int Sub(int a, int b) { return a - b; }
        public int Mul(int a, int b) { return a * b; }
        public int Div(int a, int b) { return a / b; }

        public int Ext(BinOp op, int a, int b)
        {
            if (op == null)
                throw new ArgumentNullException("op");
            return op(a, b);
        }
    }

    delegate int BinOp(int a, int b);
}
