﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.IO;
using System.Runtime.CompilerServices;
using System.Runtime.InteropServices;
using System.Text;

namespace Chapter10
{
    class Program
    {
        static void Main(string[] args)
        {
            // Call a sample method here.
        }

        static int Div(int n, int d)
        {
            if (d == 0)
                throw new ArgumentOutOfRangeException("d");

            return n / d;
        }

        static void Out1()
        {
            double root;
            if (TrySqrt(-4, out root))
                Console.WriteLine("Answer = " + root);
        }

        static void Out2()
        {
            Console.Write("Enter your age: ");
            string input = Console.ReadLine();

            int age;
            if (!int.TryParse(input, out age))
                Console.WriteLine("Error!");
            else
                Console.WriteLine("You are " + age);
        }

        static bool TrySqrt(int input, out double root)
        {
            root = 0.0;
            if (input < 0)
                return false;

            root = Math.Sqrt(input);
            return true;
        }

        static void Params()
        {
            int six = Sum(new[] { 1, 2, 3 });
            Console.WriteLine(six);

            int ten = Sum(1, 2, 3, 4);
            Console.WriteLine(ten);
        }

        static int Sum(params int[] numbers)
        {
            int sum = 0;
            foreach (var n in numbers)
                sum += n;

            return sum;
        }

        static void OptionalAndNamed1()
        {
            FileStream fs = Create(@"c:\temp\test.txt", fileOptions: FileOptions.RandomAccess);
        }

        static FileStream Create(string path, int bufferSize = 0x1000, FileOptions fileOptions = FileOptions.None)
        {
            throw new NotImplementedException("Just for illustration purposes.");
        }

        static void OptionalAndNamed2()
        {
            Console.WriteLine(Add(b: AskInt("b"), a: AskInt("a")));
        }

        static int AskInt(string name)
        {
            Console.Write(name + ": ");
            return int.Parse(Console.ReadLine());
        }

        static int Add(int a, int b)
        {
            return a + b;
        }

        static void CallerInfoAttributes1()
        {
            WriteError("Oops");
        }

        static void WriteError(string message, [CallerMemberName] string name = "", [CallerFilePath] string file = "", [CallerLineNumber] int line = -1)
        {
            Console.WriteLine("Error '{0}' occurred in '{1}' (file '{2}', line {3}).", message, name, file, line);
        }

        static void CallerInfoAttributes2()
        {
            var p = new Person("Bart", 28);
            p.Age = 29;
        }

        static void Groups()
        {
            int[] evens = Where(new[] { 1, 2, 3, 4, 5 }, Predicates.IsEven);

            foreach (var even in evens)
                Console.WriteLine(even);
        }

        static int[] Where(int[] numbers, Filter predicate)
        {
            // In fact, a more general-purpose Where method exists in LINQ.
            // It uses a generic delegate to filter elements of a generic collection.
            // For a thorough explanation, see the chapter on LINQ.
            var results = new List<int>();
            foreach (int n in numbers)
                if (predicate(n)) // delegate invocation
                    results.Add(n);
            return results.ToArray();
        }

        delegate bool Filter(int number);

        static class Predicates
        {
            public static bool IsEven(int number)
            {
                return number % 2 == 0;
            }
        }

        static void Overloads()
        {
            // Math.Abs(int)
            {
                int x = int.Parse(Console.ReadLine());
                Console.WriteLine("Abs({0}) = {1}", x, Math.Abs(x));
            }

            // Math.Abs(uint)
            {
                uint x = uint.Parse(Console.ReadLine());
                Console.WriteLine("Abs({0}) = {1}", x, Math.Abs(x));
            }
        }

        static void ExtensionMethods()
        {
            var input = Console.ReadLine();
            Console.WriteLine(input.Reverse()); // Same as using StringExtensions.Reverse(input)
        }

        static void PartialMethods()
        {
            var bart = new Person("Bart", 26);
            bart.Age = 27;
        }

        static void ExternMethods()
        {
            int size = 0;
            var user = new StringBuilder(size);
            if (!GetUserName(user, ref size))
            {
                var error = Marshal.GetLastWin32Error();
                if (error != ERROR_INSUFFICIENT_BUFFER)
                {
                    Console.WriteLine("Unexpected error: " + error);
                    return;
                }

                user = new StringBuilder(size);
                if (!GetUserName(user, ref size))
                {
                    error = Marshal.GetLastWin32Error();
                    Console.WriteLine("Unexpected error: " + error);
                    return;
                }
            }

            Console.WriteLine("Welcome, {0}", user.ToString());
        }

        const int ERROR_INSUFFICIENT_BUFFER = 122;

        [DllImport("Advapi32.dll", SetLastError = true)]
        static extern bool GetUserName(StringBuilder lpBuffer, ref int nSize);
    }

    public static class StringExtensions
    {
        public static string Reverse(this string s)
        {
            var chars = s.ToCharArray();
            Array.Reverse(chars);
            return new string(chars);
        }
    }

    // Assume this is generated code (e.g. using the LINQ to SQL designer) in a separate file.
    partial class Person : INotifyPropertyChanging, INotifyPropertyChanged
    {
        private string _name;
        private int _age;

        public Person(string name, int age)
        {
            OnCreated();
        }

        partial void OnCreated();

        public int Age
        {
            get
            {
                return _age;
            }

            set
            {
                if (_age != value)
                {
                    OnAgeChanging(value);
                    SendPropertyChanging();
                    _age = value;
                    SendPropertyChanged();
                    OnAgeChanged();
                }
            }
        }

        partial void OnAgeChanging(int value);
        partial void OnAgeChanged();

        public event PropertyChangingEventHandler PropertyChanging;
        public event PropertyChangedEventHandler PropertyChanged;

        protected virtual void SendPropertyChanging([CallerMemberName] string property = "")
        {
            var pc = PropertyChanging;
            if (pc != null)
            {
                pc(this, new PropertyChangingEventArgs(property));
            }
        }

        protected virtual void SendPropertyChanged([CallerMemberName] string property = "")
        {
            var pc = PropertyChanged;
            if (pc != null)
            {
                pc(this, new PropertyChangedEventArgs(property));
            }
        }
    }

    // Assume the developer defines custom partial methods here, outside the generated code file.
    partial class Person
    {
        partial void OnAgeChanging(int value)
        {
            if (value <= 0)
                throw new ArgumentOutOfRangeException("value");
        }

        partial void OnCreated()
        {
            Console.WriteLine("New Person created.");
        }
    }
}
