﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Reflection;
using System.Dynamic;

namespace Chapter22
{
    class Program
    {
        static void Main(string[] args)
        {
            // Call a sample method here.
        }

        static void DynamicKeyword()
        {
            // Static
            {
                string s = (string)GetObject(); // object GetObject();
                Console.WriteLine(s.ToUpper());
            }

            // Untyped - doesn't compile
            {
                //object s = GetObject(); // object GetObject();
                //Console.WriteLine(s.ToUpper() /* doesn’t compile */);
            }

            // Dynamic
            {
                dynamic s = GetObject(); // object GetObject();
                Console.WriteLine(s.ToUpper() /* does compile again */);
            }
        }

        private static object GetObject()
        {
            return "Hello";
        }

        static void Contageous()
        {
            dynamic a = 3;
            int b = 5;

            // Hover over var to see the type of c
            var c = a * b;

            Console.WriteLine(c); // This is a dynamic call too...
        }

        static void OverloadResolution()
        {
            //
            // Static typing - fully resolved at compile time
            //
            Max(3, 5);
            Max(3, 5L); // the first parameter can be implicitly converted to long
            Max(3L, 5L);

            //
            // Dynamic typing - same resolution at runtime, based on runtime types
            //
            Max((dynamic)3, 5);
            Max(3, (dynamic)5L); // same as above, but discovered by the runtime binder
            Max((dynamic)3L, (dynamic)5L);
        }

        static int Max(int a, int b)
        {
            Console.WriteLine(MethodInfo.GetCurrentMethod());
            return a > b ? a : b;
        }

        static long Max(long a, long b)
        {
            Console.WriteLine(MethodInfo.GetCurrentMethod());
            return a > b ? a : b;
        }

        static void HowItCompiles()
        {
            // Have a look at Foo.Twice in ILDASM
            new Foo().Twice(42);
        }

        class Foo
        {
            public dynamic Twice(dynamic arg)
            {
                return arg * 2;
            }
        }

        static void IronPython()
        {
            // Follow the instructions in the book to get the latest version
            // of IronPython, add references to the project, and execute the
            // code below. Don't forget to import the required namespaces too.

            /*
            {
                var py = Python.CreateEngine();
                Console.WriteLine(py.Execute("27 + 15"));
            }

            {
                var py = Python.CreateEngine();
                dynamic script = py.ImportModule("Calc");

                dynamic calc = script.GetCalc();
                Console.WriteLine(calc.Add(27, 15));
                Console.WriteLine(calc.Add(3.14, 0.0015));
                Console.WriteLine(calc.Add("Hello, ", "Python!"));
                Console.WriteLine(calc.Add(new DateTime(1983, 02, 11),
                TimeSpan.FromDays(10227)));
            }
             */
        }

        static void Internals()
        {
            // Have a look at the code for Add in ILDASM or Reflector.

            Add(27, 15); // Prints 42
            Add(DateTime.Now, TimeSpan.FromDays(7)); // Prints next week’s date and time
            Add(15, 27); // Prints 42 again
        }

        static void Add(dynamic a, dynamic b)
        {
            Console.WriteLine(a + b);
        }

        static void CustomDynamicObject_Bag()
        {
            // Classic dictionary
            var d = new Dictionary<string, object>();
            d["Name"] = "Bart";
            d["Age"] = 27;
            Console.WriteLine(d["Name"].ToString() + " is " + d["Age"].ToString());

            // Bag using dynamic
            dynamic bag = new Bag();
            bag.Name = "Bart";
            bag.Age = 27;
            Console.WriteLine(bag.Name + " is " + bag.Age);
        }

        class Bag : DynamicObject
        {
            private Dictionary<string, object> _bag = new Dictionary<string, object>();

            public override bool TryGetMember(GetMemberBinder binder, out object res)
            {
                return _bag.TryGetValue(binder.Name, out res);
            }

            public override bool TrySetMember(SetMemberBinder binder, object value)
            {
                _bag[binder.Name] = value;
                return true;
            }

            public override IEnumerable<string> GetDynamicMemberNames()
            {
                return _bag.Keys;
            }
        }

        static void Expando()
        {
            var exp = new ExpandoObject();
            // Use as generic dictionary
            var dct = (IDictionary<string, object>)exp;
            dct["Name"] = "Bart";

            // Use as dynamic bag
            var bag = (dynamic)exp;
            bag.Age = 27;

            // Use as enumerable for KeyValuePair<string, object>
            foreach (var item in dct)
                Console.WriteLine(item);
        }
    }
}
