﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Runtime.Serialization;
using System.IO;
using System.Threading.Tasks;

namespace Chapter23
{
    class Program
    {
        static string s_Name;

        static void Main(string[] args)
        {
            B();
            A();
        }

        static void A()
        {
            Console.WriteLine(s_Name.ToUpper());
        }

        static void B()
        {
            try
            {
                C();
                s_Name = "Joe";
            }
            catch (SomeException)
            {
                // Look at the generated IL code for exception handlers.
            }
        }

        static void C()
        {
            throw new SomeException();
        }

        static void DivBy0()
        {
            int n = 1, d = 0;
            int div = n / d;
            int rem = n % d;
        }

        static void Overflow()
        {
            checked
            {
                //
                // The following two are checked at compile time
                //
                // int tooHigh = int.MaxValue + 1; // add.ovf
                // int tooLow = int.MinValue - 1; // sub.ovf

                byte b = 200;
                sbyte s = (sbyte)b; // conv.ovf
            }
        }

        static void NullRef()
        {
            // Throws
            {
                string s = GetStringMaybeNull();
                string oops = s.ToUpper();
            }

            // Null-coalescing operator could help
            {
                string s = GetStringMaybeNull();
                string okay = s ?? "(empty)";
                string fine = s != null ? s.ToUpper() : "(empty)";
            }
        }

        static string GetStringMaybeNull()
        {
            return null;
        }

        static void OutOfRange()
        {
            int[] nums = new[] { 1, 2, 3 };
            for (int i = 0; i <= nums.Length; i++)
                nums[i]++;
        }

        static void InvalidCast()
        {
            object o = 42;
            string s = (string)o;
        }

        static void BrokenArrayVariance()
        {
            Uri[] uris = new[] { new Uri("http://www.bing.com") };
            object[] urisInDisguise = uris;
            urisInDisguise[1] = "http://www.google.com"; // Ow!
        }

        static void ObjectDisposed()
        {
            FileStream fs = File.OpenRead(@"..\..\Program.cs");
            using (fs)
            {
                // Do some stuff.
            }

            // This call will fail with an ObjectDisposedException as Dispose has been
            // called when the using block was exited.
            var more = fs.ReadByte();
        }

        static void OutOfMemory()
        {
            var lst = new List<byte[]>();
            while (true)
            {
                lst.Add(new byte[1024 * 1024]);
            }
        }

        static void StackOverflow_Obvious()
        {
            // May not cause an overflow on .NET 4 due to JIT optimizations.
            BlowUp();
        }

        static void BlowUp()
        {
            BlowUp();
        }

        static void StackOverflow_Data()
        {
            // Fine
            Visit(
                new TreeNode<int>
                {
                    Value = 3,
                    Left = new TreeNode<int> { Value = 1 },
                    Right = new TreeNode<int> { Value = 2 }
                }
            );

            // Oops!
            var t = new TreeNode<int> { Value = 1 };
            for (int i = 0; i < 100000; i++)
                t = new TreeNode<int> { Value = i, Left = t };

            Visit(t);
        }

        static void Visit<T>(TreeNode<T> node)
        {
            if (node.Left != null)
                Visit(node.Left);
            if (node.Right != null)
                Visit(node.Right);

            Console.WriteLine(node.Value);
        }

        class TreeNode<T>
        {
            public TreeNode<T> Left { get; set; }
            public TreeNode<T> Right { get; set; }
            public T Value { get; set; }
        }

        static void ArgumentOutOfRange()
        {
            var lst = new List<int> { 1, 2, 3 };
            int thereIsNoThirdElement = lst[3];
        }

        static void InvalidOperation()
        {
            var lst = new List<int> { 1, 2, 3 };
            foreach (var n in lst)
                lst.Add(n);
        }

        static void NotSupported()
        {
            var nums = GetNumbers();
            if (nums.MoveNext())
                Console.WriteLine(nums.Current);

            // Oops, we advanced the enumerator too far. Let’s try to reset it...
            nums.Reset(); // Throws a NotSupportedException.

            /* Here's the fix
             * 
            // Want to start with 0 again... Simply get a new iterator object.
            nums = GetNumbers();
             */

            if (nums.MoveNext())
                Console.WriteLine(nums.Current);
        }

        static IEnumerator<int> GetNumbers()
        {
            int i = 0;
            while (true)
                yield return i++;
        }

        static void Format()
        {
            int invalid = int.Parse("Hello");
        }

        static void Aggregate()
        {
            Parallel.For(0, 10, i =>
            {
                if (i % 2 == 0)
                    throw new Exception("Even number!");

                // Do meaningful work.
            });
        }
    }

    [Serializable]
    public class SomeException : Exception
    {
        public SomeException() { }
        public SomeException(string message) : base(message) { }
        public SomeException(string message, Exception inner) : base(message, inner) { }
        protected SomeException(SerializationInfo info, StreamingContext context)
            : base(info, context) { }
    }
}
