﻿using System;
using System.Reflection;

namespace Chapter25
{
    class Program
    {
        static void Main(string[] args)
        {
            // Call a sample method here.
        }

        static void AssemblyInfo()
        {
            Console.WriteLine(Assembly.GetExecutingAssembly());
            Console.WriteLine(Assembly.GetAssembly(typeof(Program)));
            Console.WriteLine(Assembly.GetAssembly(typeof(Calc)));
        }

        static void AppDomains()
        {
            var ad = AppDomain.CreateDomain("App1");
            Console.WriteLine(ad.Id + " " + ad.FriendlyName);
            ad.ExecuteAssembly(@"c:\temp\app1.exe"); // TODO for reader: put a managed assembly in this path
            AppDomain.Unload(ad);
        }

        static void AppDomain_CreateInstance()
        {
            var ad = AppDomain.CreateDomain("AddIn");

            dynamic calc = ad.CreateInstanceFromAndUnwrap("Calc.dll", "Calc"); // TODO for reader: copy Chapter25_Calc\bin\Debug\calc.dll to Chapter25\bin\Debug
            Console.WriteLine(calc.Add(1, 2));

            AppDomain.Unload(ad);

            foreach (var asm in AppDomain.CurrentDomain.GetAssemblies())
                Console.WriteLine(asm);
        }

        static void Remoting_MarshalByRef()
        {
            var ad = AppDomain.CreateDomain("AddIn");

            dynamic calc = ad.CreateInstanceFromAndUnwrap("Calc.dll", "Calc2"); // TODO for reader: copy Chapter25_Calc\bin\Debug\calc.dll to Chapter25\bin\Debug
            Console.WriteLine(calc.Add(1, 2));

            AppDomain.Unload(ad);
        }

        static void Remoting_Serializable()
        {
            var ad = AppDomain.CreateDomain("AddIn");

            dynamic calc = ad.CreateInstanceFromAndUnwrap("Calc.dll", "Calc3"); // TODO for reader: copy Chapter25_Calc\bin\Debug\calc.dll to Chapter25\bin\Debug
            Console.WriteLine(calc.Add(1, 2));

            AppDomain.Unload(ad);
        }

        static void AddIn_CommonInterface()
        {
            var ad = AppDomain.CreateDomain("AddIn");

            var calc = (IBinaryOperation)ad.CreateInstanceFromAndUnwrap("Calc.dll", "Pythagorean"); // TODO for reader: copy Chapter25_Calc\bin\Debug\calc.dll to Chapter25\bin\Debug
            Console.WriteLine(calc.Compute(3, 4));
        }
    }
}
